/***************************************************************************
                          drawingenvironment.cpp  -  description
                             -------------------
    begin                : Sun Apr 14 2002
    copyright            : (C) 2002 by Andrew Bibb
    email                : buyer242299@aol.com
    source               : http://lx-viewer.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, version 2 of the License, or any later  * 
 *   version.                                                              *
 *                                                                         *
 ***************************************************************************/

//	Functions to control the drawing environment such as
//	ortho
//	snap
//	modelspace
//	paperspace

#include "drawingenvironment.h"
#include "vectorize.h"
#include "resource.h"
#include "infomessage.h"

#include "xpm/buttonon.xpm"
#include "xpm/buttonoff.xpm"

extern PAD_DWGHDR adhd;

///////////////////////////////////////////////////////////////////////////////////////////
//					The lighted push button
///////////////////////////////////////////////////////////////////////////////////////////
LX_LightedButton::LX_LightedButton(QWidget* parent, const char* name):QPushButton(parent,name)
{
	boff = QPixmap(buttonoff);
	bon = QPixmap(buttonon);

	setToggleButton(TRUE);
	setToggleType(Toggle);
	setIconSet(QIconSet(boff));
	
	connect(this,SIGNAL(toggled(bool)),this,SLOT(buttonDown(bool)));
}
	
// slots (private) for this button
//	following sets the button without emiting a toggled signal
void LX_LightedButton::buttonDown(bool oo)
{
	oo? displayon() : displayoff();
	return;
}

//	following 2 slots do not emit the toggled signal
void LX_LightedButton::displayon()
{
	setDown(true);
	setIconSet(QIconSet(bon));
}

void LX_LightedButton::displayoff()
{
	setDown(false);
	setIconSet(QIconSet(boff));
}

///////////////////////////////////////////////////////////////////////////////////////////////
//						The settings bar
///////////////////////////////////////////////////////////////////////////////////////////////

//	Neither Grid nor Snap appear to be saved with the drawing.  I believe this in not correct
//	but at the same time I can't find an error in the code below.  I
//	Need to do some research to find out how to save these values.  Intellicad seems to be able
//	to save them with the drawing.

LX_SettingsBar::LX_SettingsBar(QWidget* parent, const char* name, WFlags f):QWidget(parent,name,f),
	snap_string(tr("SNAP")),grid_string(tr("GRID")),ortho_string(tr("ORTHO")),isnap_string(tr("ISNAP")),
	model_string(tr("MODEL")),paper_string(tr("PAPER")),tile_string(tr("TILE"))
{
	//	member data needed by other routines
	lastpslayout = 1;
	
  //	graphics primitive sizes
	const int gap = 5;
	const int buttonwidth = 68;
	const int barheight = 19;
	const int combowidth = 160;
	const int border = 1;
	int x = border;
	int y = border;

	setMinimumSize(2*border+5*gap+5*buttonwidth+combowidth,2*border+barheight);
	setMaximumSize(2*border+5*gap+5*buttonwidth+combowidth,2*border+barheight);

	//	buttons
	snapbutton = new LX_LightedButton(this);
		snapbutton->setGeometry(x,y,buttonwidth,barheight);
		snapbutton->setText(snap_string);
		snapbutton->setDisabled(true);
		x += gap + buttonwidth;	
	gridbutton = new LX_LightedButton(this);
		gridbutton->setGeometry(x,y,buttonwidth,barheight);
		gridbutton->setText(grid_string);
		gridbutton->setDisabled(true);
		x += gap + buttonwidth;
	orthobutton = new LX_LightedButton(this);
		orthobutton->setGeometry(x,y,buttonwidth,barheight);
		orthobutton->setText(ortho_string);
		orthobutton->setDisabled(true);
		x += gap + buttonwidth;
	isnapbutton = new QPushButton(this);
		isnapbutton->setGeometry(x,y,buttonwidth,barheight);
		isnapbutton->setText(isnap_string);
		isnapbutton->setDisabled(true);
		x += gap + buttonwidth;
	tilebutton = new LX_LightedButton(this);
		tilebutton->setGeometry(x,y,buttonwidth,barheight);
		tilebutton->setText(tile_string);
		tilebutton->setDisabled(true);
		x += gap + buttonwidth;
	layoutbox = new QComboBox(this);
		layoutbox->setGeometry(x,y,combowidth,barheight);

		layoutbox->setEditable(false);
		layoutbox->setDisabled(true);

	//	connect signals and slots
	connect(snapbutton,SIGNAL(toggled(bool)),this,SLOT(snapclicked(bool)));
	connect(gridbutton,SIGNAL(toggled(bool)),this,SLOT(gridclicked(bool)));
	connect(orthobutton,SIGNAL(toggled(bool)),this,SLOT(orthoclicked(bool)));
	connect(tilebutton,SIGNAL(clicked()),this,SLOT(tileclicked()));
	connect(layoutbox,SIGNAL(activated(int)),this,SLOT(layoutpicked(int)));

	//	now put in some "Whats This" text
	QWhatsThis::add(snapbutton,tr("Button to toggle snapmode on and off"));
	QWhatsThis::add(gridbutton,tr("Button to toggle gridmode on and off"));
	QWhatsThis::add(orthobutton,tr("Button to toggle orthomode on and off"));
	QWhatsThis::add(isnapbutton,tr("Button to set the entity snap modes"));
	QWhatsThis::add(tilebutton,tr("Button to toggle between Model Space and Paper Space views"));
	QWhatsThis::add(layoutbox,tr("Select the layout to display - Layouts are only valid for versions of ACAD 2000 and above.  "
		"If the file version is R14 or below the file version is listed in this box"));
}

/////////////////////////////////////////////////////
//	SettingsBar Private Functions
////////////////////////////////////////////////////
//	make sure the paperspace block contains at least one viewport.  If it
//	does return true.
bool LX_SettingsBar::validPSBlock(AD_OBJHANDLE h)
{
	AD_BLKH blkh;
	AD_ENT aden;
	AD_ENT_HDR adenhd;
	int psVpCount = 0;

	adSeekBlockheader(filehandle,h,&blkh);
	adStartEntityGet(blkh.entitylist);
	while (adGetEntity(blkh.entitylist,&adenhd,&aden))
	{
		if (adenhd.enttype == AD_ENT_VIEWPORT) psVpCount++;
	}

	return (psVpCount > 0);
}

void LX_SettingsBar::findlayouts()
{
	AD_OBJ adob;
	AD_OBJ_HDR adobhd;
	AD_BLKH blkh;

	layoutbox->clear();	
	if (filehandle && (adFileVersionLoaded(filehandle) < AD_ACAD2000))
	{
		short ver = adFileVersionLoaded(filehandle);
		switch (ver)
		{
			case AD_ACAD25:
				layoutbox->insertItem(tr("Release 2.5 Drawing"));
				break;
			case AD_ACAD26:
				layoutbox->insertItem(tr("Release 2.6 Drawing"));
				break;
			case AD_ACAD9:
				layoutbox->insertItem(tr("Release 9 Drawing"));
				break;
			case AD_ACAD10:
				layoutbox->insertItem(tr("Release 10 Drawing"));
				break;
			case AD_ACAD11:
				layoutbox->insertItem(tr("Release 11/12 Drawing"));
				break;
			case AD_ACAD13:
				layoutbox->insertItem(tr("Release 13 Drawing"));
				break;
			case AD_ACAD14:
				layoutbox->insertItem(tr("Release 14 Drawing"));
				break;
			default:
				layoutbox->insertItem(tr("Unknown Drawing Version"));
		}
		return;
	}
	adStartObjectGet(filehandle);	
	while (adGetObject(filehandle,&adobhd,&adob))
	{
		if (adobhd.objtype == adLayoutObjtype(filehandle) &&
			adSeekBlockheader(filehandle,adob.layout.pspaceblockrec, &blkh))		
		{
			if (strcmp(adob.layout.name, "Model") == 0)	//	Model space always first
				layoutbox->insertItem(adob.layout.name,0);
			else if (validPSBlock(blkh.objhandle))
			{
				layoutbox->insertItem(adob.layout.name, -1);
			}	
		}	//	double if at start
	}	//	while loop	
}
	

//////////////////////////////////////////////////////
//	SettingsBar Public functions
//////////////////////////////////////////////////////
void LX_SettingsBar::initializebar(AD_DB_HANDLE dh)
{
	filehandle = dh;
	adhd = adHeaderPointer(filehandle);
	findlayouts();
	lastpslayout = 1;

	snapMode(adhd->snapmode);
	gridMode(adhd->gridmode);
	orthoMode(adhd->orthomode);

	if (adhd->tilemode == 1)	//	calling layoutpicked() will cause a revectorization
	{
		layoutbox->setCurrentItem(0);
		layoutpicked(0);
	}
	else
	{
		layoutbox->setCurrentItem(layoutbox->count() > 1? 1:0);
		layoutpicked(1);
	}	
	
	tilebutton->setDisabled(false);
	layoutbox->setDisabled(false);
}

void LX_SettingsBar::snapMode(short m){m == 1? snapbutton->setOn(true):snapbutton->setOn(false);}
void LX_SettingsBar::gridMode(short m){m == 1? gridbutton->setOn(true):gridbutton->setOn(false);}
void LX_SettingsBar::orthoMode(short m){m == 1? orthobutton->setOn(true):orthobutton->setOn(false);}

// Function below is slightly mis-named.  It does not return an AD_OBJHANDLE, rather
//	it sets the AD_OBJHANDLE pointed to by *layouthandle to the appropriate value
void LX_SettingsBar::getLayoutHandle(AD_OBJHANDLE* layouthandle)
{
	AD_OBJ adob;
	AD_OBJ_HDR adobhd;
	AD_BLKH blkh;

	adHanclear(*layouthandle);	//	default is null handle
	
	if (layoutbox->currentItem() == 0) return; 	//	model space so return

	adStartObjectGet(filehandle);	
	while (adGetObject(filehandle,&adobhd,&adob))
	{
		if (adobhd.objtype == adLayoutObjtype(filehandle) &&
			adSeekBlockheader(filehandle,adob.layout.pspaceblockrec, &blkh))
		{
			adHancpy (*layouthandle, adob.layout.pspaceblockrec);		
			if (QString(adob.layout.name) == layoutbox->currentText()) break;
		}	
	
	}	
}


//////////////////////////////////////////////////////
//	Settings Bar Public slots
//////////////////////////////////////////////////////
void LX_SettingsBar::snapclicked(bool c)
{
	adhd = adHeaderPointer(filehandle);
	adhd->snapmode = c? 1:0;
}

void LX_SettingsBar::gridclicked(bool c)
{
	adhd = adHeaderPointer(filehandle);
	adhd->gridmode = c? 1:0;
}

void LX_SettingsBar::orthoclicked(bool c)
{
	adhd = adHeaderPointer(filehandle);
	adhd->orthomode = c? 1:0;
}

void LX_SettingsBar::isnapclicked(){}

void LX_SettingsBar::tileclicked()
{
	if (tilebutton->isOn())
	{
		layoutbox->setCurrentItem(0);
		layoutpicked(0);
	}
	else
	{
		if (layoutbox->count() > 1) layoutbox->setCurrentItem(lastpslayout);
		layoutpicked(1);
	}
}
	
void LX_SettingsBar::layoutpicked(int pos)
{
	AD_OBJHANDLE psHandle;
	AD_OBJHANDLE psBlockHandle;	

	adhd = adHeaderPointer(filehandle);
	adHanclear(psHandle);
	
  if (pos ==  0) //	model space always first in the popup list
	{
		if (!tilebutton->isDown()) tilebutton->toggle();	
		tilebutton->setText(model_string);
		if (adhd->tilemode != 1) adhd->tilemode = 1;
	}
	else 	//	either straight PS (R14 and below) or some PS layout (A2K and above)
	{
		adGetBlockHandle(filehandle,psHandle,AD_PAPERSPACE_HANDLE);
		if (!validPSBlock(psHandle))					//	return if no valid PS block handle
		{
			DisplayMessage(NO_PS_VIEWPORTS);	//	display message
			tilebutton->toggle();
			tilebutton->setText(model_string);
			return;
		}		   	
 		if (tilebutton->isDown()) tilebutton->toggle();
		tilebutton->setText(paper_string);	
		if (adhd->tilemode != 0) adhd->tilemode = 0;
		lastpslayout = pos;
	}
	
//	either model space was selected or a valid paperspace viewport exists and was
//	selected so change to the correct space and revectorize
	getLayoutHandle(&psBlockHandle);
	emit commandString(GetResourceString(TILE_MODE) +  QString::number(adhd->tilemode,10));
	vectorize(filehandle,psBlockHandle);
}
		

