/***************************************************************************
                          svgexport.cpp  -  description
                             -------------------
    begin                : Sat Nov 23 2002
    copyright            : (C) 2002 by Andrew Bibb
    email                : buyer242299@aol.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, version 2 of the License, or any later  * 
 *   version.                                                              *
 *                                                                         *
 ***************************************************************************/

/*	Code for Scalable Vector Graphics (a subset of XML) output
To use this class first construct a new svg output member (remember this is derived from a qFile class)
Then set the color using the setColor() member function createRoot will read this to set the background rect.
Then create the root element using createRoot()
Then startWrite() to open the file and seed it with the root information
Call setsvgfuns (located inside OD_Interface)
Then vectorize the drawing, and finally call endWrite() to close the open file*/

#include "svgexport.h"

//	constructor for svgExport class - derived from a QFile class
svgExport::svgExport(const QString name) : QFile(name)
{
	
	content = "<?xml version=\"1.0\" standalone=\"no\"?>";
	pub_id	=	"-//W3C//DTD SVG 1.0//EN";
 	sys_id 	= "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd";
	nspace	=	"http://www.w3.org/2000/svg";
	s_red 	= "FF";
	s_green	= "FF";
	s_blue	= "FF";
	linewidth = 1;	//	no changes to this are implemented yet
	rootopen = false;
	grpopen	= false;
	curpath = QString();
	ptcntr = 0;
	nline_at = 10;
}

//	public functions
//	function to open the file and write the head and root data
//	return true on success, false on error
bool svgExport::startWrite()
{
	if (open(IO_WriteOnly))
	{
		os.setDevice(this);
		QDateTime qdt = QDateTime::currentDateTime();
		
	  os << content;
		os << "\n<!DOCTYPE svg PUBLIC \"" << pub_id << "\" \"" << sys_id << "\">";
		os << "\n<!--LX-Viewer SVG Export-->";
		os << "\n<!--Created: " << qdt.toString(Qt::TextDate) << "-->";
	  if (rootopen) os << rootelem;
		return true;
	}
	return false;
}

//	function to close the file and end the writing process
//	return true on success, false on error
bool svgExport::endWrite()
{
	if (!curpath.isNull()) os << curpath << "\" />";
	if (grpopen) 	os << "\n</g>";
	if (rootopen) os << "\n</g>\n</svg>";

	close();
	return true;
}

//	function to create the root element plus some other header
//	type tags.  w and h are the width and height values
void svgExport::createRoot(long w, long h)
{
	QString s_width = QString::number(w-1);
	QString s_height = QString::number(h-1);
	rootopen = false;
	grpopen	= false;
	
 	rootelem = "\n<svg xmlns=\"";
	rootelem.append(nspace);
	rootelem.append("\" width=\"100%\" height=\"100%\" viewBox=\"0 0 ");
	rootelem.append(s_width);
	rootelem.append(" ");
	rootelem.append(s_height);
	rootelem.append("\" >");
	
	//	set the background color to match our drawingarea screen
	//	no need for stroke or line width settings
	rootelem.append("\n");
	rootelem.append("<rect fill=\"#" + s_red + s_green + s_blue + "\" ");
	rootelem.append("x=\"0\" y=\"0\" width=\"" + s_width + "\" height=\"" + s_height + "\" />");

	//	create an outter g element for constant stroke settings
	rootelem.append("\n");
	rootelem.append("<g stroke-linejoin=\"miter\" fill-rule=\"evenodd\" >");

	rootopen = true;
}

//	function to add a moveto element to the tree
void svgExport::moveTo(const long& x, const long& y)
{
	ptcntr++;
		
	if(curpath.isNull()) curpath = QString("\n<path d=\"");
	curpath.append(QString(" M") + QString::number(x) + "," + QString::number(y));
	if(ptcntr % nline_at == 0) curpath.append("\n");
	startx = x;
	starty = y;
}

//	function to add a lineto element to the tree
void svgExport::lineTo(const long& x, const long& y)
{
	if (startx == x && starty == y) return;

	ptcntr++;
	
	if(curpath.isNull()) curpath = QString("\n<path d=\" M") + QString::number(startx) + "," + QString::number(starty);
	curpath.append(QString(" L") + QString::number(x) + "," + QString::number(y));
	if(ptcntr % nline_at == 0) curpath.append("\n");
	startx = x;
	starty = y;
}

//	function to set the color values
//	each color change sets a new group
void svgExport::setColor(const QColor& c)
{
	const int base(16);
	QString s;

	//	need to pad these hex values into 2 character strings
	s = QString::number(qRed(c.rgb()),base);
	s_red = (s.length() == 1 ? QString("0") + s : s);
	s = QString::number(qGreen(c.rgb()),base);
	s_green = (s.length() == 1 ? QString("0") + s : s);
	s = QString::number(qBlue(c.rgb()),base);
	s_blue = (s.length() == 1 ? QString("0") + s : s);

	if (rootopen)
	{
		if (grpopen)
		{
			if (!curpath.isNull()) os << curpath << "\" />";
			os << "\n</g>";
			curpath = QString();
		}
		os << "\n<g ";
		os << "fill=\"none\" ";
		os << "stroke=\"#" << s_red << s_green << s_blue << "\" ";
		os << "stroke-width=\"" << QString::number(linewidth) << "\" >";
		ptcntr = 0;
		grpopen = true;
	}
}

//	function to paint and fill a polygon
void svgExport::paintPoly(const QPointArray& a)
{
	int x(0);
	int y(0);
	int prevx = -1;
	int prevy = -1;
	int pointcounter(0);
	QString curpoly = QString();
	QString pointlist = QString();
	const QString fillstuff(" fill=\"" + QString("#") + s_red + s_green + s_blue + "\"");

	for (unsigned int i = 0; i < a.size(); i++)
	{
		a.point(i,&x,&y);
		if (x != prevx || y != prevy)
		{
			pointlist.append(QString::number(x) + ",");
			pointlist.append(QString::number(y) + " ");
			prevx = x;
			prevy = y;
			pointcounter++;
		}		
	}

	if (pointcounter < 2) return;
	if (pointcounter == 2)	curpoly = QString("\n<polyline points=\"");
	if (pointcounter > 2)  	curpoly = QString("\n<polygon" + fillstuff + " points=\"");	

 	curpoly.append(pointlist);
	curpoly.append("\" />");
	os << curpoly;
}




